@@ -150,8 +150,14 @@ _X11TransConnectDisplay (
     char* address = addrbuf;
     XtransConnInfo trans_conn = NULL;	/* transport connection object */
     int connect_stat;
+#if defined(LOCALCONN) || defined(TCPCONN)
+    Bool reset_hostname = False;	/* Reset hostname? */
+#endif
 #ifdef LOCALCONN
     struct utsname sys;
+# ifdef UNIXCONN    
+    Bool try_unix_socket = False;	/* Try unix if local fails */
+# endif    
 #endif
 #ifdef TCPCONN
     char *tcphostname = NULL;		/* A place to save hostname pointer */
@@ -222,19 +228,28 @@ _X11TransConnectDisplay (
 
 #ifdef LOCALCONN
     /* check if phostname == localnodename AND protocol not specified */
-    if (!pprotocol && phostname && uname(&sys) >= 0 &&
+    if (!pprotocol && (!phostname || (phostname && uname(&sys) >= 0 &&
 	!strncmp(phostname, sys.nodename, 
 	(strlen(sys.nodename) < strlen(phostname) ? 
-	strlen(phostname) : strlen(sys.nodename))))
+	 strlen(phostname) : strlen(sys.nodename))))))
     {
-#ifdef TCPCONN
 	/*
 	 * We'll first attempt to connect using the local transport.  If
+	 * that fails, we'll try again using the Unix socket transport.  If
 	 * this fails (which is the case if sshd X protocol forwarding is
 	 * being used), retry using tcp and this hostname.
 	 */
-	tcphostname = copystring(phostname, strlen(phostname));
+#ifdef UNIXCONN
+	try_unix_socket = True;
 #endif
+#ifdef TCPCONN
+	if (phostname)
+	    tcphostname = copystring(phostname, strlen(phostname));
+	else
+	    tcphostname = copystring("localhost", 9);
+#endif
+	if (!phostname)
+	    reset_hostname = True;
 	Xfree (phostname);
 	phostname = copystring ("unix", 4);
     }
@@ -288,7 +303,11 @@ _X11TransConnectDisplay (
 
 #if defined(TCPCONN) || defined(UNIXCONN) || defined(LOCALCONN) || defined(MNX_TCPCONN) || defined(OS2PIPECONN)
     if (!pprotocol) {
+#ifdef HAVE_LAUNCHD
+	if (!phostname || phostname[0]=='/') {
+#else
 	if (!phostname) {
+#endif
 #if defined(UNIXCONN) || defined(LOCALCONN) || defined(OS2PIPECONN)
 	    pprotocol = copystring ("local", 5);
 #if defined(TCPCONN)
@@ -419,11 +438,29 @@ _X11TransConnectDisplay (
      *
      *     [host] : [:] dpy . scr \0
      */
+#if defined(LOCALCONN) || defined(TCPCONN)
+    /*
+     *  If we computed the host name, get rid of it so that
+     *  XDisplayString() and XDisplayName() agree.
+     */ 
+    if (reset_hostname) {
+	Xfree (phostname);
+	phostname = NULL;
+    }
+#endif
     len = ((phostname ? strlen(phostname) : 0) + 1 + (dnet ? 1 : 0) +
 	   strlen(pdpynum) + 1 + (pscrnum ? strlen(pscrnum) : 1) + 1);
     *fullnamep = (char *) Xmalloc (len);
     if (!*fullnamep) goto bad;
 
+#ifdef HAVE_LAUNCHD
+    if (phostname && strlen(phostname) > 11 && !strncmp(phostname, "/tmp/launch", 11)) 
+    	sprintf (*fullnamep, "%s%s%d",
+	     (phostname ? phostname : ""),
+	     (dnet ? "::" : ":"),
+	     idisplay);
+    else
+#endif
     sprintf (*fullnamep, "%s%s%d.%d",
 	     (phostname ? phostname : ""),
 	     (dnet ? "::" : ":"),
@@ -454,11 +491,21 @@ _X11TransConnectDisplay (
     if (phostname) Xfree (phostname);
     if (address && address != addrbuf) { Xfree(address); address = addrbuf; }
 
+#if defined(LOCALCONN) && defined(UNIXCONN)
+    if (try_unix_socket) {
+	pprotocol = copystring ("unix", 4);
+	phostname = NULL;
+	try_unix_socket = False; /* Do this only once */
+	goto connect;
+    }
+#endif
+    
 #if defined(TCPCONN)
     if (tcphostname) {
 	pprotocol = copystring("tcp", 3);
 	phostname = tcphostname;
 	tcphostname = NULL;
+	reset_hostname = True;
 	goto connect;
     }
 #endif
@@ -712,11 +759,11 @@ void XSetAuthorization (name, namelen, data, datalen)
  * Create a credential that we can send to the X server.
  */
 static int
-auth_ezencode(servername, window, cred_out, len)
-        char           *servername;
-        int             window;
-	char	       *cred_out;
-        int            *len;
+auth_ezencode(
+    char           *servername,
+    int             window,
+    char	   *cred_out,
+    int            *len)
 {
         AUTH           *a;
         XDR             xdr;
