@@ -17,21 +17,28 @@ static void _XCBLockDisplay(Display *dpy)
 {
     if(dpy->xcb->lock_fns.lock_display)
 	dpy->xcb->lock_fns.lock_display(dpy);
-    xcb_xlib_lock(dpy->xcb->connection);
-    _XGetXCBBuffer(dpy);
+    if(!dpy->lock || dpy->lock->locking_level == 0)
+	xcb_xlib_lock(dpy->xcb->connection);
+    if(!(dpy->flags & XlibDisplayIOError))
+	_XGetXCBBuffer(dpy);
 }
 
+/* XXX: If you change this function, update _XReply's copy of its guts! */
 static void _XCBUnlockDisplay(Display *dpy)
 {
-    _XPutXCBBuffer(dpy);
-    assert(dpy->xcb->partial_request == 0);
-    assert(xcb_get_request_sent(dpy->xcb->connection) == dpy->request);
+    if(!(dpy->flags & XlibDisplayIOError))
+    {
+	_XPutXCBBuffer(dpy);
+	assert(dpy->xcb->partial_request == 0);
+	assert(xcb_get_request_sent(dpy->xcb->connection) == dpy->request);
 
-    /* Traditional Xlib does this in _XSend; see the Xlib/XCB version
-     * of that function for why we do it here instead. */
-    _XSetSeqSyncFunction(dpy);
+	/* Traditional Xlib does this in _XSend; see the Xlib/XCB version
+	 * of that function for why we do it here instead. */
+	_XSetSeqSyncFunction(dpy);
+    }
 
-    xcb_xlib_unlock(dpy->xcb->connection);
+    if(!dpy->lock || dpy->lock->locking_level == 0)
+	xcb_xlib_unlock(dpy->xcb->connection);
     if(dpy->xcb->lock_fns.unlock_display)
 	dpy->xcb->lock_fns.unlock_display(dpy);
 }
@@ -47,6 +54,14 @@ int _XCBInitDisplayLock(Display *dpy)
     return 1;
 }
 
+void _XCBShutdownDisplayLock(Display *dpy)
+{
+    if(dpy->lock_fns) {
+	Xfree((char *)dpy->lock_fns);
+	dpy->lock_fns = NULL;
+    }
+}
+
 void _XGetXCBBuffer(Display *dpy)
 {
     static const xReq dummy_request;
